/* ////////////////////////////////////////////////////////////////////////////
 * File:        winstl/pid_sequence.hpp
 *
 * Purpose:     Process Id sequence class.
 *
 * Created:     24th June 2005
 * Updated:     11th September 2005
 *
 * Home:        http://stlsoft.org/
 *
 * Copyright (c) 2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/** \file winstl/pid_sequence.hpp Process Id sequence class. */

#ifndef WINSTL_INCL_WINSTL_HPP_PID_SEQUENCE
#define WINSTL_INCL_WINSTL_HPP_PID_SEQUENCE

#ifndef STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_WINSTL_HPP_PID_SEQUENCE_MAJOR       1
# define WINSTL_VER_WINSTL_HPP_PID_SEQUENCE_MINOR       1
# define WINSTL_VER_WINSTL_HPP_PID_SEQUENCE_REVISION    4
# define WINSTL_VER_WINSTL_HPP_PID_SEQUENCE_EDIT        11
#endif /* !STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Compatibility
 */

/*
[Incompatibilies-start]
STLSOFT_COMPILER_IS_COMO:
[Incompatibilies-end]
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include <winstl.h>
#endif /* !WINSTL_INCL_H_WINSTL */
#ifndef WINSTL_INCL_WINSTL_H_PROCESS_STATUS
# include <winstl/process_status.h>
#endif /* !WINSTL_INCL_WINSTL_H_PROCESS_STATUS */
#ifdef __STLSOFT_CF_EXCEPTION_SUPPORT
# ifndef WINSTL_INCL_WINSTL_HPP_EXCEPTIONS
#  include <winstl/exceptions.hpp>
# endif /* !WINSTL_INCL_WINSTL_HPP_EXCEPTIONS */
#endif /* __STLSOFT_CF_EXCEPTION_SUPPORT */
#ifndef WINSTL_INCL_WINSTL_HPP_PROCESSHEAP_ALLOCATOR
# include <winstl/processheap_allocator.hpp>
#endif /* !WINSTL_INCL_WINSTL_HPP_PROCESSHEAP_ALLOCATOR */
#ifndef WINSTL_INCL_WINSTL_HPP_SYSTEM_VERSION
# include <winstl/system_version.hpp>
#endif /* !WINSTL_INCL_WINSTL_HPP_SYSTEM_VERSION */
#ifndef STLSOFT_INCL_STLSOFT_HPP_AUTO_BUFFER
# include <stlsoft/auto_buffer.hpp>
#endif /* !STLSOFT_INCL_STLSOFT_HPP_AUTO_BUFFER */
#ifndef STLSOFT_INCL_STLSOFT_ALGORITHMS_HPP_POD
# include <stlsoft/algorithms/pod.hpp>
#endif /* !STLSOFT_INCL_STLSOFT_ALGORITHMS_HPP_POD */
#if !defined(STLSOFT_UNITTEST)
# include <algorithm>
#endif /* !STLSOFT_UNITTEST */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

namespace stlsoft
{

namespace winstl_project
{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Process Id sequence
class pid_sequence
{
public:
    /// The value type
    typedef DWORD                               value_type;
    /// The allocator type
    typedef processheap_allocator<value_type>   allocator_type; 
    /// The class type
    typedef pid_sequence                        class_type;
    /// The non-mutating (const) pointer type
    typedef value_type const                    *const_pointer;
    /// The non-mutating (const) iterator type
    typedef value_type const                    *const_iterator;
    /// The non-mutating (const) reference type
    typedef value_type const                    &const_reference;
    /// The size type
    typedef ws_size_t                           size_type;
    /// The difference type
    typedef ws_ptrdiff_t                        difference_type;

    enum
    {
            elideIdle   =   0x0001
        ,   elideSystem =   0x0002
        ,   sort        =   0x0004
    };

public:
    /// Constructs a sequence from the current processes in the host system
    ss_explicit_k pid_sequence(ws_uint32_t flags = elideIdle | elideSystem);
    /// Copies the contents of the sequence 
    pid_sequence(class_type const &rhs);
    /// Releases the storage associated with the process id list
    ~pid_sequence() stlsoft_throw_0();

public:
    /// Begins the iteration
    ///
    /// \return An iterator representing the start of the sequence
    const_iterator  begin() const;
    /// Ends the iteration
    ///
    /// \return An iterator representing the end of the sequence
    const_iterator  end() const;

    /// Returns a reference to the element at the given index
    ///
    /// \param index The required index. Behaviour is undefined, if not within the range [0, size())
    const_reference operator [](size_type index) const;

public:
    /// Indicates whether the sequence is empty
    ws_bool_t   empty() const;
    /// Returns the number of identifiers in the sequence
    size_type   size() const;

public:
    static value_type   idleProcessId()
    {
        return 0;
    }
    static value_type   systemProcessId()
    {
        ws_uint32_t major   =   system_version::major();
        ws_uint32_t minor   =   system_version::minor();

        if(4 == major)
        {
            return 2;   // NT 4
        }
        else
        {
            if( 5 == major &&
                0 == minor)
            {
                return 8; // Win2K
            }
            else if(5 == major &&
                    1 == minor)
            {
                return 4; // WinXP
            }
            else
            {
                return 4; // Longhorn and above - this value is a guess!!
            }
        }
    }

private:
    typedef stlsoft_ns_qual(auto_buffer)<   value_type
                                        ,   allocator_type
                                        ,   128
                                        >       buffer_type;

    buffer_type m_pids;

private:
    class_type &operator =(class_type const &);
};

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST
# include "./unittest/pid_sequence_unittest_.h"
#endif /* STLSOFT_UNITTEST */

/* /////////////////////////////////////////////////////////////////////////////
 * Implementation
 */

#ifndef STLSOFT_DOCUMENTATION_SKIP_SECTION

inline pid_sequence::pid_sequence(ws_uint32_t flags)
    : m_pids(buffer_type::internal_size())
{
    DWORD   cbReturned;

    for(;;)
    {
        if(!::EnumProcesses(&m_pids[0], sizeof(value_type) * m_pids.size(), &cbReturned))
        {
#ifdef __STLSOFT_CF_EXCEPTION_SUPPORT
            throw windows_exception("Failed to enumerate processes", ::GetLastError());
#else /* ? __STLSOFT_CF_EXCEPTION_SUPPORT */
            m_pids.resize(0);

            break;
#endif /* __STLSOFT_CF_EXCEPTION_SUPPORT */
        }
        else
        {
            const size_t    n   =   cbReturned / sizeof(value_type);

            if(n < m_pids.size())
            {
                m_pids.resize(n);

                break;
            }
            else
            {
                if(!m_pids.resize(2 * m_pids.size()))
                {
                    // This will only ever be executed when compiled in the
                    // absence of throwing bad_alloc on memory exhaustion
                    m_pids.resize(0);

                    break;
                }
            }
        }
    }

    if(flags & (elideIdle | elideSystem))
    {
        value_type  *begin      =   &*m_pids.begin();
        value_type  *end        =   &*m_pids.end();
        value_type  *pIdle      =   (flags & elideIdle) ? stlsoft_ns_qual_std(find)(begin, end, idleProcessId()) : end;
        value_type  *pSystem    =   (flags & elideSystem) ? stlsoft_ns_qual_std(find)(begin, end, systemProcessId()) : end;

        // Optimise for the special case where idle is [0] and system is [1]
        if( end != pIdle &&
            end != pSystem &&
            pSystem == pIdle + 1)
        {
            pod_move(pSystem + 1, end, begin);
            m_pids.resize(m_pids.size() - 2);
        }
        else
        {
            if(end != pIdle)
            {
                pod_move(pIdle + 1, end, pIdle);
                m_pids.resize(m_pids.size() - 1);
            }
            if(end != pSystem)
            {
                pod_move(pSystem + 1, end, pSystem);
                m_pids.resize(m_pids.size() - 1);
            }
        }
    }

    if(flags & sort)
    {
        stlsoft_ns_qual_std(sort)(m_pids.begin(), m_pids.end());
    }
}

inline pid_sequence::pid_sequence(pid_sequence const &rhs)
    : m_pids(rhs.m_pids.size())
{
    stlsoft_ns_qual_std(copy)(rhs.m_pids.begin(), rhs.m_pids.end(), m_pids.begin());
}

inline pid_sequence::~pid_sequence() stlsoft_throw_0()
{}

inline pid_sequence::const_iterator pid_sequence::begin() const
{
    return &*m_pids.begin();
}

inline pid_sequence::const_iterator pid_sequence::end() const
{
    return &*m_pids.end();
}

inline pid_sequence::const_reference pid_sequence::operator [](pid_sequence::size_type index) const
{
    winstl_message_assert("Index out of range", index < size());

    return m_pids[index];
}

inline ws_bool_t pid_sequence::empty() const
{
    return m_pids.empty();
}

inline pid_sequence::size_type pid_sequence::size() const
{
    return m_pids.size();
}

#endif /* !STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

} // namespace winstl_project

} // namespace stlsoft

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !WINSTL_INCL_WINSTL_HPP_PID_SEQUENCE */

/* ////////////////////////////////////////////////////////////////////////// */
